import backend.controllers as ctrls
from backend.models import Card, FedCard, DssUser, DataColumn
from backend.models.card import CardType
from backend.models.data_column import FedDataColumn, DataColumnLabelType
from backend.schemas import CardSchema, DataColumnSchema, NotProvided
from backend.utils.misc import clean_value
from marshmallow import ValidationError
from werkzeug.exceptions import NotFound, BadRequest
import functools
from typing import List, Dict, Any
from collections import OrderedDict

EMPTY_CONTENT = {}


def card_data_bundler(func):
    @functools.wraps(func)
    def wrapper(content_raw_data):
        if not len(content_raw_data):
            return EMPTY_CONTENT
        try:
            return func(content_raw_data)
        except Exception as err:
            error_message = str(err)
            return {"error": error_message}

    return wrapper


class CardController(ctrls.BaseController):
    def get_all(self):
        return Card.query.all()

    def get_by_id(self, card_id: int, ignore_if_not_found: bool = False):
        if card_id is None:
            if not ignore_if_not_found:
                raise ValidationError(message={"card_id": ["Field may not be null."]})
            return None
        card: Card = Card.query.get(card_id)
        if card is None and not ignore_if_not_found:
            raise NotFound("Card not found.")
        return card

    def get_data_columns(self, card_id: int) -> List[DataColumn]:
        _ = self.get_by_id(card_id)

        return DataColumn.query.filter_by(card_id=card_id).all()

    def get_many(self, card_ids: List[int], ignore_if_not_found: bool = False):
        if card_ids is None:
            if not ignore_if_not_found:
                raise ValidationError(message={"card_ids": ["Field may not be null."]})
            return []
        card_ids = list(set(card_ids))
        cards: List[Card] = Card.query.filter(Card.id.in_(card_ids)).all()
        if (len(cards) < len(card_ids)) and not ignore_if_not_found:
            raise NotFound("Some cards were not found.")
        return cards

    def feed_one(self, card_id: int, connection_data: Dict[str, Any], dss_user: DssUser):
        """Fill card with data coming from python function execution or directly jsonpaths"""
        card = self.get_by_id(card_id)

        error = None

        if connection_data is None:
            card_data = {}
        else:
            try:
                card_data = self.process_card_content(card.id, connection_data)
            except Exception as err:
                card_data = {}
                error = f"Error when generating card content: {str(err)}."

        if error is None and card_data is None:
            return None

        fed_card = FedCard(
            card=card,
            view=card.view,
            data=card_data,
            error=error,
            is_unpinned=ctrls.user_config_ctrl.is_unpinned(card_id, dss_user),
            is_hidden=ctrls.user_config_ctrl.is_hidden(card_id, dss_user),
            position=0,
        )
        return fed_card

    def export(self, card_id: int):
        card = self.get_by_id(card_id)

        fields_to_export = [
            "name",
            "type",
            "priority",
            "data_columns",
        ]
        return CardSchema(only=tuple(fields_to_export)).dump(card)

    def create(
        self,
        view_id: int,
        name: str,
        type: CardType,
        priority: int,
        data_columns: List[DataColumn] = NotProvided,
    ):
        _ = ctrls.view_ctrl.get_by_id(view_id)

        new_card = CardSchema().load(
            {
                "view_id": view_id,
                "name": name,
                "type": type,
                "data_columns": data_columns,
                "priority": priority,
            },
            session=self.db.session,
        )

        self.db.session.add(new_card)
        self.db.session.commit()

        return self.get_by_id(new_card.id)

    def create_many(self, view_id: int, cards: List[dict]) -> List[Card]:
        new_cards = []
        for card in cards:
            new_card = self.create(view_id=view_id, **card)
            new_cards.append(new_card)
        return new_cards

    def update(
        self,
        card_id: int,
        name: str = NotProvided,
        type: CardType = NotProvided,
        data_columns: List[Dict] = NotProvided,
        priority: int = NotProvided,
    ):
        card = self.get_by_id(card_id)

        new_card = CardSchema().load(
            {
                "type": type,
                "name": name,
                "data_columns": data_columns,
                "priority": priority,
            },
            instance=card,
            session=self.db.session,
            partial=True,
        )

        self.db.session.add(new_card)
        self.db.session.commit()

        return self.get_by_id(card_id)

    def update_card_priority(self, card_id: int, new_index: int, old_index: int) -> bool:
        card = self.get_by_id(card_id)

        new_priority = new_index
        old_priority = old_index

        sup_priority, inf_priority = list(sorted([new_priority, old_priority]))
        incr = +1 if new_priority < old_priority else -1

        cards_to_update = Card.query.filter(Card.priority <= inf_priority, Card.priority >= sup_priority).all()

        for r in cards_to_update:
            r.priority = r.priority + incr

        card.priority = new_priority

        self.db.session.add_all(cards_to_update)
        self.db.session.commit()

        return True

    def delete(self, card_id):
        card_to_delete = self.get_by_id(card_id)

        self.db.session.delete(card_to_delete)
        self.db.session.commit()

        is_card_deleted = self.get_by_id(card_id, ignore_if_not_found=True) == None
        if not is_card_deleted:
            raise BadRequest("Could not delete card.")

        return True

    def update_data_columns(self, card_id: int, data_columns: List[Dict]) -> List[DataColumn]:
        card_to_update = self.get_by_id(card_id)

        data_columns_deserialized = DataColumnSchema().load(
            data_columns,
            many=True,
            session=self.db.session,
        )

        for i, data_col in enumerate(data_columns_deserialized):
            data_col.priority = i

        card_to_update.data_columns = data_columns_deserialized

        self.db.session.add(card_to_update)
        self.db.session.commit()

    def is_fed_card_empty(self, fed_card):
        return not fed_card.data

    def test_create_card(self, card_id: int, dss_user: DssUser):
        card = self.get_by_id(card_id)

        connection_data = ctrls.view_ctrl.fetch_test_keyword_data(card.view_id)
        fed_card = self.feed_one(card_id, connection_data=connection_data, dss_user=dss_user)

        return fed_card

    def feed_one_data_column(self, data_column: DataColumn, connection_data: Any) -> FedDataColumn:
        if not data_column.name in connection_data:
            return None
        if data_column.label_type == DataColumnLabelType.CONSTANT:
            label_value = data_column.label
        else:
            if data_column.label in connection_data:
                label_value = connection_data[data_column.label]
            else:
                label_value = data_column.label
        return FedDataColumn(data_column=data_column, value=connection_data[data_column.name], label_value=label_value)

    def feed_data_columns(self, data_columns: List[DataColumn], connection_data: Any) -> List[FedDataColumn]:
        fed_data_columns = []
        for data_col in data_columns:
            new_fed_data_column = self.feed_one_data_column(data_col, connection_data)
            if new_fed_data_column:
                fed_data_columns.append(new_fed_data_column)
        return fed_data_columns

    def process_card_content(self, card_id, connection_data):
        _mapping_card_type_to_processing = {
            CardType.TEXT: self.bundle_data_for_text,
            CardType.IFRAME: self.bundle_data_for_iframe,
            CardType.LIST: self.bundle_data_for_list,
            CardType.LINK: self.bundle_data_for_link,
            CardType.IMAGE: self.bundle_data_for_image,
            CardType.MARKDOWN: self.bundle_data_for_markdown,
            CardType.KPI: self.bundle_data_for_list,
        }

        card = self.get_by_id(card_id)
        data_columns: List[DataColumn] = card.data_columns
        data_columns.sort(key=lambda data_col: data_col.priority)

        content_processor = _mapping_card_type_to_processing.get(card.type)
        if content_processor is None:
            raise BadRequest(f"Type {card.type} has no content processor.")
        fed_data_columns = self.feed_data_columns(data_columns, connection_data)

        card_content = content_processor(fed_data_columns)

        return card_content

    @staticmethod
    @card_data_bundler
    def bundle_data_for_text(fed_data_columns: List[FedDataColumn]):
        """Format the card content to display the input values as a text card. Such values can be `list`, `float`, `int`, `str`."""
        unescaped_values = [clean_value(c.value) for c in fed_data_columns]
        formatted_text = "\n".join(unescaped_values)

        if not formatted_text:
            return EMPTY_CONTENT

        card_content = {"text": formatted_text}
        return card_content

    @staticmethod
    @card_data_bundler
    def bundle_data_for_markdown(fed_data_columns: List[FedDataColumn]):
        card_data = []
        for data_col in fed_data_columns:
            unescaped_value = clean_value(data_col.value)
            if unescaped_value:
                card_data.append({"label": data_col.label_value, "value": unescaped_value})

        if not len(card_data):
            return EMPTY_CONTENT

        card_content = {"markdown": card_data}
        return card_content

    @staticmethod
    @card_data_bundler
    def bundle_data_for_iframe(fed_data_columns: List[FedDataColumn]):
        iframe_data = fed_data_columns[0]

        if not iframe_data.value:
            return EMPTY_CONTENT

        card_content = {"iframe": {"label": iframe_data.label_value, "value": iframe_data.value}}
        return card_content

    @staticmethod
    @card_data_bundler
    def bundle_data_for_list(fed_data_columns: List[FedDataColumn]):
        """Format the card content to display the input values as list card."""
        card_data = []
        for data_col in fed_data_columns:
            value = data_col.value if data_col.value else "-"
            card_data.append({"label": data_col.label_value, "value": value})

        card_content = {"list": card_data}
        return card_content

    @staticmethod
    @card_data_bundler
    def bundle_data_for_link(fed_data_columns: List[FedDataColumn]):
        card_data = []
        for data_col in fed_data_columns:
            if data_col.value:
                card_data.append({"label": data_col.label_value, "value": data_col.value})

        if not len(card_data):
            return EMPTY_CONTENT

        card_content = {"links": card_data}
        return card_content

    @staticmethod
    @card_data_bundler
    def bundle_data_for_image(fed_data_columns: List[FedDataColumn]):
        """Format the card content to display the input values as a text card. Such values can be `list`, `float`, `int`, `str`."""
        image_data = fed_data_columns[0]

        if not image_data:
            return EMPTY_CONTENT

        card_content = {"image": {"label": image_data.label_value, "value": image_data.value}}
        return card_content
