import copy
import logging
import datetime
import json

from backend.models import Connection, Keyword
from backend.models.connection import MultipleMatchesMode
from backend.schemas import ConnectionSchema
from marshmallow import ValidationError
import backend.controllers as ctrls
from backend.schemas.base_schema import NotProvided
from werkzeug.exceptions import NotFound, BadRequest
from typing import List, Dict, Any, Union, Optional
from backend.utils.sql_executor import (
    select_columns_from_dataset,
    select_first_from_dataset_where_columns_are_values,
    get_rows_count,
)
from backend.utils.misc import get_index_or_none
from backend.utils.dataiku_api import dataiku_api

logger = logging.getLogger(__name__)
BIG_DATASET_THRESHOLD = 500_000


class ConnectionController(ctrls.BaseController):
    def get_by_id(self, connection_id: int, ignore_if_not_found: bool = False) -> Connection:
        if connection_id is None:
            if not ignore_if_not_found:
                raise ValidationError(message={"connection_id": ["Field may not be null."]})
            return None
        connection: Connection = Connection.query.get(connection_id)
        if connection is None and not ignore_if_not_found:
            raise NotFound("Connection not found.")
        return connection

    def get_by_id_and_set_default_keyword_if_needed(
        self, connection_id: int, ignore_if_not_found: bool = False
    ) -> Connection:
        connection = self.get_by_id(connection_id, ignore_if_not_found)
        if not connection.test_keyword.primary:
            self.set_default_test_keyword_if_ready(connection_id)
        return connection

    def get_all(self) -> List[Connection]:
        return Connection.query.all()

    def get_active(self, connections: List[Connection]) -> List[Connection]:
        return [conn for conn in connections if self.is_active(conn.id)]

    def get_all_active(self) -> List[Connection]:
        connections = self.get_all()
        active_connections = self.get_active(connections=connections)
        return active_connections

    def is_active(self, connection_id: int) -> bool:
        connection = self.get_by_id(connection_id)
        return ctrls.view_ctrl.is_active(connection.view_id)

    def export(self, connection_id: int) -> Dict[str, Any]:
        connection = self.get_by_id(connection_id)

        fields_to_export = [
            "dataset_name",
            "keywords_column_name",
            "multiple_matches_mode",
            "secondary_column_name",
        ]
        connection_as_dict = ConnectionSchema(only=tuple(fields_to_export)).dump(connection)
        return connection_as_dict

    def create(
        self,
        view_id: int,
        dataset_name: str = NotProvided,
        keywords_column_name: str = NotProvided,
        multiple_matches_mode: MultipleMatchesMode = NotProvided,
        secondary_column_name: str = NotProvided,
    ) -> Connection:
        new_connection = ConnectionSchema().load(
            {
                "view_id": view_id,
                "dataset_name": dataset_name,
                "keywords_column_name": keywords_column_name,
                "multiple_matches_mode": multiple_matches_mode,
                "secondary_column_name": secondary_column_name,
            },
            session=self.db.session,
        )
        default_test_keyword = self.get_default_test_keyword(new_connection)
        new_connection.test_keyword = default_test_keyword
        self.db.session.add(new_connection)
        self.db.session.commit()

        return self.get_by_id(new_connection.id)

    def update(
        self,
        connection_id: int,
        dataset_name: str = NotProvided,
        keywords_column_name: str = NotProvided,
        multiple_matches_mode: MultipleMatchesMode = NotProvided,
        secondary_column_name: str = NotProvided,
        test_keyword: Dict = NotProvided,
    ) -> Connection:
        connection = self.get_by_id(connection_id)
        old_connection = copy.deepcopy(connection)

        updated_connection: Connection = ConnectionSchema().load(
            {
                "dataset_name": dataset_name,
                "keywords_column_name": keywords_column_name,
                "multiple_matches_mode": multiple_matches_mode,
                "secondary_column_name": secondary_column_name,
                "test_keyword": test_keyword,
            },
            instance=connection,
            session=self.db.session,
            partial=("id", "view_id"),
        )

        if multiple_matches_mode and multiple_matches_mode == MultipleMatchesMode.KEEP_FIRST:
            updated_connection.secondary_column_name = None

        self.db.session.add(updated_connection)
        self.db.session.commit()

        should_reset_default_keyword = self.should_reset_default_keyword(
            old_connection=old_connection, new_connection=updated_connection
        )

        if should_reset_default_keyword and not test_keyword:
            self.set_default_test_keyword_if_ready(connection_id)

        return self.get_by_id(connection_id)

    def should_reset_default_keyword(self, old_connection: Connection, new_connection: Connection) -> bool:
        return (
            not new_connection.test_keyword
            or new_connection.dataset_name != old_connection.dataset_name
            or new_connection.keywords_column_name != old_connection.keywords_column_name
            or new_connection.multiple_matches_mode != old_connection.multiple_matches_mode
            or new_connection.secondary_column_name != old_connection.secondary_column_name
        )

    def is_connection_ready(self, connection: Connection) -> bool:
        return (
            connection.dataset_name
            and connection.keywords_column_name
            and (
                connection.multiple_matches_mode == MultipleMatchesMode.KEEP_FIRST
                or (
                    connection.multiple_matches_mode == MultipleMatchesMode.SECONDARY_COLUMN
                    and connection.secondary_column_name
                )
            )
        )

    def populate_keywords_search(
        self,
        connection_id: int,
        match_string: str = "",
        batch_size: int = 100,
        batches_offset: int = 0,
    ):
        keywords = self.fetch_keywords(
            connection_id=connection_id,
            like_string=match_string,
            limit=batch_size,
            order_by=True,
            batch_offset=batches_offset,
        )

        return keywords

    def delete(self, connection_id):
        connection_to_delete = self.get_by_id(connection_id)

        self.db.session.delete(connection_to_delete)
        self.db.session.commit()

        is_connection_deleted = self.get_by_id(connection_id, ignore_if_not_found=True) == None
        if not is_connection_deleted:
            raise BadRequest("Could not delete connection.")

        return True

    def set_default_test_keyword_if_ready(self, connection_id: int):
        connection = self.get_by_id(connection_id)
        if not self.is_connection_ready(connection):
            return

        connection.test_keyword = self.get_default_test_keyword(connection)

        self.db.session.add(connection)
        self.db.session.commit()

    def get_default_test_keyword(self, connection: Connection) -> str:
        test_keyword = self.fetch_keywords_with_connection(connection, limit=1)
        return test_keyword[0] if len(test_keyword) != 0 else Keyword(primary="")

    def fetch_keywords(
        self,
        connection_id: int,
        limit: int = None,
        like_string: Union[str, List[str]] = "",
        order_by: bool = False,
        batch_offset: int = 0,
    ) -> List[Keyword]:
        connection = self.get_by_id(connection_id)
        return self.fetch_keywords_with_connection(
            connection=connection,
            limit=limit,
            like_string=like_string,
            order_by=order_by,
            batch_offset=batch_offset,
        )

    def fetch_keywords_with_connection(
        self,
        connection: Connection,
        limit: int = None,
        like_string: Union[str, List[str]] = "",
        order_by: bool = False,
        batch_offset: int = 0,
    ):
        if not self.is_connection_ready(connection):
            return []
        try:
            secondary_column_name = connection.secondary_column_name if MultipleMatchesMode.SECONDARY_COLUMN else None
            keywords_list = self.fetch_keywords_ready(
                dataset_name=connection.dataset_name,
                keywords_column_name=connection.keywords_column_name,
                secondary_column_name=secondary_column_name,
                like_string=like_string,
                limit=limit,
                order_by=order_by,
                batch_offset=batch_offset,
            )
        except Exception as err:
            logger.error(f"Error while fetching keywords for connection {connection.id}.")
            logger.exception(err)
            keywords_list = []
        return keywords_list

    def fetch_keywords_ready(
        self,
        dataset_name: str,
        keywords_column_name: str,
        secondary_column_name: Optional[str] = None,
        limit: int = None,
        like_string: Union[str, List[str]] = "",
        order_by: bool = False,
        batch_offset: int = 0,
    ) -> List[Keyword]:
        
        params = dict(
            dataset_name=dataset_name,
            columns_name=[keywords_column_name],
            format_="iter",
            keep_none=[False],
            like_string=[like_string],
            limit=limit,
            order_by=order_by,
            batch_offset=batch_offset,
        )

        if secondary_column_name:
            params["columns_name"].append(secondary_column_name)
            params["keep_none"].append(True)
            params["like_string"].append("")

        if dataiku_api.is_dataset_sql(dataset_name):
            rows_count = get_rows_count(dataset_name)
            is_too_big_for_distinct = rows_count > BIG_DATASET_THRESHOLD
            params["distinct"] = not is_too_big_for_distinct
            query_result = select_columns_from_dataset(**params)
        else:
            query_result = dataiku_api.select_columns_from_dataset(**params)
        keywords_set = set()

        for kw in query_result:
            new_kw = Keyword(primary=kw[0], secondary=get_index_or_none(kw, 1))
            keywords_set.add(new_kw)

        keywords_list = list(keywords_set)
        return keywords_list


    def clean_data_for_serialization(self, data: dict) -> dict:
        DEFAULT_DATE_FORMAT = "%Y-%m-%d"
        DEFAULT_TIME_FORMAT = "%H:%M:%S"
        DEFAULT_DATETIME_FORMAT = f"{DEFAULT_DATE_FORMAT} {DEFAULT_TIME_FORMAT}"

        new_data = {}

        is_date = lambda value: isinstance(value, datetime.date) or (
            isinstance(value, datetime.datetime)
            and (value.hour == 0 and value.minute == 0 and value.second == 0 and value.microsecond == 0)
        )
        is_time = lambda value: isinstance(value, datetime.time) or (
            isinstance(value, datetime.datetime) and (value.year == 1900 and value.month == 1 and value.day == 1)
        )

        def is_jsonable(value):
            try:
                json.dumps(value)
                return True
            except (TypeError, OverflowError):
                return False

        for key, value in data.items():
            if is_date(value):
                new_data[key] = value.strftime(DEFAULT_DATE_FORMAT)
            elif is_time(value):
                new_data[key] = value.strftime(DEFAULT_TIME_FORMAT)
            elif isinstance(value, datetime.datetime):
                new_data[key] = value.strftime(DEFAULT_DATETIME_FORMAT)
            elif not is_jsonable(value):
                new_data[key] = str(value)
            else:
                new_data[key] = value
        return new_data

    def fetch_data_ready(self, dataset_name: str, keywords_column_name: str, keyword: Keyword, secondary_column_name: Optional[str] = None) -> Dict:

        params = dict(
            dataset_name=dataset_name,
            filter_columns=[keywords_column_name],
            filter_values=[keyword.primary],
            format_="iter",
        )

        if secondary_column_name:
            params["filter_columns"].append(secondary_column_name)
            params["filter_values"].append(keyword.secondary)

        if dataiku_api.is_dataset_sql(dataset_name):
            query_result = select_first_from_dataset_where_columns_are_values(**params)
        else:
            query_result = dataiku_api.select_first_from_dataset_where_columns_are_values(**params)
        
        if query_result is None:
            logger.info("Could not find any results running the query.")
            return {}
        
        connection_data_deserialized = list(query_result)
        n_results = len(connection_data_deserialized)
        if n_results == 0:
            return {}
        connection_data_first = connection_data_deserialized[0]
        dataset_columns = dataiku_api.list_columns(dataset_name)

        connection_data_schema = [c["name"] for c in dataset_columns]
        return dict(zip(connection_data_schema, connection_data_first))

    def fetch_data(self, connection_id: int, keyword: Keyword):
        connection = self.get_by_id(connection_id)
        if not self.is_connection_ready(connection):
            return {}

        try:
            secondary_column_name = connection.secondary_column_name if MultipleMatchesMode.SECONDARY_COLUMN else None
            connection_data = self.fetch_data_ready(
                dataset_name=connection.dataset_name,
                keywords_column_name=connection.keywords_column_name,
                secondary_column_name=secondary_column_name,
                keyword=keyword,
            )
        except Exception as err:
            logger.error(f"Error while fetching data for connection {connection_id}.")
            logger.exception(err)
            connection_data = {}

        connection_data_ready = self.clean_data_for_serialization(connection_data)
        return connection_data_ready

    def fetch_test_data(self, connection_id: int):
        connection = self.get_by_id(connection_id)

        test_connection_data = self.fetch_data(connection_id=connection_id, keyword=connection.test_keyword)

        return test_connection_data

    def get_test_keyword(self, connection_id: int):
        connection = self.get_by_id(connection_id)
        test_keyword: Keyword = copy.deepcopy(connection.test_keyword)

        if connection.multiple_matches_mode == MultipleMatchesMode.SECONDARY_COLUMN:
            test_keyword.secondary = getattr(test_keyword, "secondary", "")
        else:
            if hasattr(test_keyword, "secondary"):
                delattr(test_keyword, "secondary")
        return test_keyword
