import logging
import backend.controllers as ctrls
from backend.models import UserConfig, DssUser
from backend.models.user_config import UserConfigParam
from werkzeug.exceptions import BadRequest
import json
from typing import List, Any

from backend.utils.misc import NotProvided

logger = logging.getLogger(__name__)


class UserConfigController(ctrls.BaseController):
    def get_user_config(self, param: UserConfigParam, dss_user: DssUser, create_if_none: bool = False) -> Any:
        user_config = UserConfig.query.filter_by(dss_user_id=dss_user.login, param=param).first()

        if user_config is None:
            if create_if_none:
                user_config = UserConfig(param=param, param_value=json.dumps({}), dss_user_id=dss_user.login)
                self.db.session.add(user_config)
                self.db.session.commit()
            else:
                return None
        try:
            user_param_config: Any = json.loads(user_config.param_value)
        except Exception as err:
            logger.error(
                f"The UserConfig of user {dss_user.login} is not json formatted. Value is: {user_config}\nOriginal error: {err}"
            )
            raise BadRequest("Something went wront when fetching the user configurations.")
        return user_param_config

    def update_user_config(self, param: UserConfigParam, param_value: Any, dss_user: DssUser) -> bool:
        user_config = UserConfig.query.filter_by(dss_user_id=dss_user.login, param=param).first()

        if user_config is None:
            return True

        try:
            user_param_config: str = json.dumps(param_value)
        except Exception as err:
            logger.error(f"Could not dump param_value {param_value} for user {dss_user.login}.\nOriginal error: {err}")
            raise BadRequest("Something went wront when dumping the user configurations.")

        user_config.param_value = user_param_config

        self.db.session.add(user_config)
        self.db.session.commit()

    def get_cards_ordering_for_view(self, view_id: int, dss_user: DssUser) -> List[str]:
        user_ordered_cards_by_view = self.get_user_config(param=UserConfigParam.CARDS_ORDERING, dss_user=dss_user)

        if user_ordered_cards_by_view:
            return user_ordered_cards_by_view.get(view_id, [])

        return []

    def toggle_card_visibility(self, card_id: int, dss_user: DssUser):
        user_hidden_cards = self.get_user_config(
            param=UserConfigParam.CARD_VISIBILITY,
            dss_user=dss_user,
            create_if_none=True,
        ).get("cards_id", [])

        is_now_hidden = not (card_id in user_hidden_cards)
        if is_now_hidden:
            user_hidden_cards.append(card_id)
        else:
            user_hidden_cards.remove(card_id)

        self.update_user_config(
            param=UserConfigParam.CARD_VISIBILITY,
            param_value={"cards_id": user_hidden_cards},
            dss_user=dss_user,
        )
        return is_now_hidden

    def toggle_card_unpin(self, card_id: int, dss_user: DssUser):
        user_unpin_cards = self.get_user_config(
            param=UserConfigParam.CARD_UNPIN, dss_user=dss_user, create_if_none=True
        ).get("cards_id", [])

        is_now_unpin = not (card_id in user_unpin_cards)
        if is_now_unpin:
            user_unpin_cards.append(card_id)
        else:
            user_unpin_cards.remove(card_id)

        self.update_user_config(
            param=UserConfigParam.CARD_UNPIN,
            param_value={"cards_id": user_unpin_cards},
            dss_user=dss_user,
        )

        return True

    def update_cards_ordering(self, dss_user: DssUser, view_id: str, cards_ordering: List[str] = NotProvided) -> bool:
        user_ordered_cards = self.get_user_config(
            param=UserConfigParam.CARDS_ORDERING, dss_user=dss_user, create_if_none=True
        )

        user_ordered_cards[view_id] = cards_ordering if cards_ordering != NotProvided else []

        self.update_user_config(
            param=UserConfigParam.CARDS_ORDERING,
            param_value=user_ordered_cards,
            dss_user=dss_user,
        )

        return True

    def is_hidden(self, card_id, dss_user):
        hidden_cards_id = self.get_user_config(param=UserConfigParam.CARD_VISIBILITY, dss_user=dss_user)
        if not hidden_cards_id:
            return False

        hidden_cards_id = hidden_cards_id["cards_id"]
        return card_id in hidden_cards_id

    def is_unpinned(self, card_id, dss_user):
        unpinned_cards_id = self.get_user_config(param=UserConfigParam.CARD_UNPIN, dss_user=dss_user)
        if not unpinned_cards_id:
            return False
        unpinned_cards_id = unpinned_cards_id["cards_id"]
        return card_id in unpinned_cards_id
