from werkzeug.exceptions import NotFound
from typing import Dict, Any, Optional, Union, List
from flask import current_app
import logging
import os
import json
from enum import Enum
from typing_extensions import Literal
import pandas as pd
from backend.utils.dataset_csv_reader import DataikuStreamedHttpUTF8CSVReader
from backend.utils.misc import escape_string_double
from dataikuapi.dss.dataset import DSSDataset
from dataikuapi.dss.flow import DSSFlowZone

logger = logging.getLogger(__name__)

try:
    import dataiku
    from dataiku.customwebapp import get_webapp_config, get_plugin_config
except ImportError:
    logger.debug("Working outside of Dataiku.")


logger = logging.getLogger(__name__)

NO_TRACKING_CONFIG = {"udr_mode": "NO"}
ADVISOR_LIGHT_BLUE = "#7DF9FF"


class DatasetStatus(Enum):
    SUCCESS = 1
    EMPTY = 2
    ERROR = 3

class DataikuApi:
    def __init__(self):
        self._webapp_config = None
        self._plugin_config = None
        self._instanse_info = None
        self._default_project = None
        try:
            import dataiku
        except:
            raise ImportError(
                "cannot import dataiku library. Add dataiky python library to the path (see /python-lib/backend/README.md)"
            )

        self._client = dataiku.api_client()
        self.set_instance_info()

    def setup(
        self,
        webapp_config: Dict,
        plugin_config: Dict,
        default_project_key: Optional[str] = None,
    ):

        self._webapp_config = webapp_config
        self._default_project_key = default_project_key
        self._plugin_config = plugin_config

    @property
    def default_project(self):
        try:
            return self.client.get_default_project()
        except Exception as err:
            if self._default_project_key:
                return self.client.get_project(self._default_project_key)
            else:
                raise Exception("Please define the default project before using it.")

    @property
    def client(self):
        if self._client is None:
            raise Exception("Please set the client before using it.")
        else:
            return self._client

    @client.setter
    def client(self, c: Any):
        raise Exception("If working outside of Dataiku, Client can only be set through the function setup()")
    
    def set_webapp_config(self, webapp_id: str, new_config: Dict):
        webapp_settings = self.default_project.get_webapp(webapp_id).get_settings()

        current_saved_config: Dict = webapp_settings.get_raw()['config']
        current_env_config: Dict = get_webapp_config()

        current_saved_config.update(new_config)
        current_env_config.update(new_config)

        webapp_settings.save()
        os.environ["DKU_CUSTOM_WEBAPP_CONFIG"] = json.dumps(current_env_config)

    def get_project(self, project_id=None):
        if project_id is None:
            return self.default_project
        else:
            return self.client.get_project(project_id)

    def is_dataset_sql_supported(self, dataset_type: str):
        return dataset_type in DSSDataset._SQL_TYPES

    def list_datasets(self, project_id=None, connection=None):
        project = self.get_project(project_id)
        datasets = project.list_datasets()
        if connection:
            datasets = [d for d in datasets if d["params"].get("connection") == connection]
        return datasets

    def list_sql_datasets(self, project_id=None):
        datasets = self.list_datasets(project_id=project_id)
        sql_datasets = [d for d in datasets if self.is_dataset_sql_supported(d.get("type"))]
        return sql_datasets

    def list_datasets_summary(self, project_id=None, connection=None):
        datasets = self.list_datasets(project_id, connection)
        return [
            {"name": ds["name"], "type": ds["type"], "is_sql": self.is_dataset_sql_supported(ds["type"])}
            for ds in datasets
        ]

    def list_sql_datasets_summary(self, project_id=None):
        sql_datasets = self.list_sql_datasets(project_id)
        return [{"name": ds["name"], "type": ds["type"]} for ds in sql_datasets]

    def list_columns(self, dataset_name: str, project_id: Optional[str] = None):
        project = self.get_project(project_id)
        dataset = project.get_dataset(dataset_name)
        try:
            schema = dataset.get_schema()
        except Exception as err:
            raise NotFound("Dataset not found.")
        return schema["columns"]
    
    def get_dataset_status(self, dataset_name: str, project_id: Optional[str] = None):
        project = self.get_project(project_id)
        dataset = project.get_dataset(dataset_name)
        status = DatasetStatus.SUCCESS
        try:
            _ = next(dataset.iter_rows())
        except StopIteration:
            status = DatasetStatus.EMPTY
        except Exception as err:
            status = DatasetStatus.ERROR

        return status

    @property
    def in_dataiku(self):
        return os.getenv("FLASK_ENV") == "dss"

    @property
    def webapp_config(self):
        return self._webapp_config or get_webapp_config()

    @property
    def plugin_config(self):
        return self._plugin_config or get_plugin_config()

    @property
    def public_plugin_config(self):
        public_config = ["activate_tracking", "csp_headers_allowlist", "samesite_headers_allowlist"]
        plugin_config = self.plugin_config
        public_plugin_config = {k: v for k, v in plugin_config.items() if k in public_config}
        return public_plugin_config

    def is_dev_instance(self):
        try:
            import dataiku

            return dataiku.get_dss_settings()["devInstance"]
        except Exception:
            return False

    def is_user_admin(self):
        try:
            _ = self.client.get_general_settings()
            return True
        except Exception as err:
            return False

    def get_tracking_config(self, headers):
        if not self.in_dataiku:
            return NO_TRACKING_CONFIG

        import dataiku

        is_user_admin = self.is_user_admin()
        plugin_config = self.plugin_config
        udr_mode = "DEFAULT"
        if is_user_admin:
            general_settings = self.client.get_general_settings().get_raw()
            if "udr" in general_settings:
                udr_mode = "DEFAULT" if general_settings["udr"] else "NO"
            if "udrMode" in general_settings:
                udr_mode = general_settings["udrMode"]
        else:
            general_settings = {}
            udr_mode = "DEFAULT"

        if not plugin_config.get("activate_tracking"):
            udr_mode = "NO"

        dss_settings: Dict = dataiku.get_dss_settings()
        is_dev_instance = dss_settings.get("devInstance", False)

        is_tracking_enabled = udr_mode != "NO" and not is_dev_instance

        if not is_tracking_enabled:
            return {"udr_mode": udr_mode}

        try:
            instance_info = self.get_instance_info()
        except Exception:
            instance_info = {}

        if is_user_admin:
            licensing_status = self.client.get_licensing_status()
            try:
                dss_license_kind = licensing_status["base"]["licenseContent"]["licenseKind"]
            except KeyError:
                dss_license_kind = None
        else:
            licensing_status = {}
            dss_license_kind = None

        try:
            user_info: Dict = self.get_user_info(headers)
        except Exception as _:
            auth_info: Dict = self.get_auth_info(headers)
            user_info = {
                "login": auth_info.get("associatedDSSUser"),
                "userProfile": auth_info.get("userProfile"),
            }

        is_dev = os.getenv("FLASK_ENV") == "development"

        tracking_config = {
            "udr_mode": udr_mode,
            "dip_instance_id": instance_info.get("dipInstanceId"),
            "install_id": instance_info.get("installId"),
            "dss_license_kind": dss_license_kind,
            "bkd_distrib": dss_settings.get("distrib"),
            "bkd_distrib_version": dss_settings.get("distribVersion"),
            "node_type": instance_info.get("nodeType"),
            "dss_version": instance_info.get("dssVersion"),
            "install_id": instance_info.get("installId"),
            "is_dev_instance": is_dev_instance or is_dev,
            "user_info": user_info,
            "web_app_id": current_app.config.get("WEB_APP_ID"),
        }

        return tracking_config

    def get_auth_info(self, headers):
        try:
            return self.client.get_auth_info_from_browser_headers(headers)
        except Exception as err:
            return self.client.get_auth_info()

    def get_user_info(self, headers):
        auth_info = self.get_auth_info(headers)
        return self.client.get_user(auth_info["authIdentifier"]).get_settings().get_raw()

    def get_connection_info(self, connection_name: str) -> Dict:
        return self.client.get_connection(connection_name).get_info()

    def execute_sql(self, *args, **kwargs):
        return self.client.sql_query(*args, **kwargs)

    def set_instance_info(self):
        try:
            self._instanse_info = self.client.get_instance_info().raw
        except Exception:
            self.instance_info = {}

    def get_instance_info(self) -> Any:
        if self._instanse_info == None:
            self.set_instance_info()
        return self._instanse_info

    def build_like_for_formula(self, column_name: str, like_str: str) -> str:
        return f'contains(toLowercase(toString(strval("{escape_string_double(column_name)}"))), "{str(escape_string_double(like_str)).lower()}")'

    def build_like_filter_string(self, columns: List[str], like_string: Union[str, List[str]]) -> str:
        full_filter = "1==1"
        if isinstance(like_string, str):
            if len(like_string) == 0:
                return full_filter
            like_string = escape_string_double(like_string=like_string).lower()
            all_filters = [self.build_like_for_formula(c, like_string) for c in columns]
            full_filter = f"and({', '.join(all_filters)})"

        elif isinstance(like_string, list):
            if all([len(ls) == 0 for ls in like_string]):
                return full_filter
            if len(like_string) != len(columns):
                raise Exception("like_string should have the same length as columns_name.")
            all_filters = []
            for i, ls in enumerate(like_string):
                if len(ls):
                    ls_ready = escape_string_double(like_string=ls).lower()
                    all_filters.append(self.build_like_for_formula(columns[i], ls_ready))
            full_filter = f"and({', '.join(all_filters)})"

        return full_filter

    def build_keep_none_filter_string(self, columns: List[str], keep_none: Union[bool, List[bool]] = False) -> str:
        full_keep_none_filter = "1==1"
        if isinstance(keep_none, bool):
            if not keep_none:
                all_keep_none_str = [f'isNotNull(strval("{escape_string_double(c)}"))' for c in columns]
                full_keep_none_filter = f"and({', '.join(all_keep_none_str)})"

        elif isinstance(keep_none, list):
            if len(keep_none) != len(columns):
                raise Exception("keep_none should have the same length as columns_name.")
            all_keep_none_str = []
            for i, kn in enumerate(keep_none):
                if columns[i] and not kn:
                    all_keep_none_str.append(f'isNotNull(strval("{escape_string_double(columns[i])}"))')
            full_keep_none_filter = f"and({', '.join(all_keep_none_str)})"

        return full_keep_none_filter

    def get_dataset_data(
        self, project_key: str, dataset_name: str, columns: List[str], sampling: str, filter: str
    ) -> Dict:
        path = f"projects/{project_key}/datasets/{dataset_name}/data"
        body = {"format": "json", "columns": columns, "sampling": sampling, "filter": filter}
        self._client._perform_json(method="POST", path=path, body=body)

    def select_columns_from_dataset(
        self,
        dataset_name: str,
        columns_name: List[str],
        format_: Literal["dataframe", "iter"] = "dataframe",
        keep_none: Union[bool, List[bool]] = False,
        like_string: Union[str, List[str]] = "",
        limit: int = None,
        distinct: bool = False,
        order_by: bool = False,
        batch_offset: int = 0,
    ) -> Union[pd.DataFrame, iter]:
        dataset = self.default_project.get_dataset(dataset_name)
        project_key = self.default_project.project_key

        like_filter_str = self.build_like_filter_string(columns_name, like_string)
        keep_none_filter_str = self.build_keep_none_filter_string(columns_name, keep_none)

        full_filter = f"and({like_filter_str}, {keep_none_filter_str})"

        csv_stream = self.client._perform_raw(
            "POST",
            f"/projects/{project_key}/datasets/{dataset_name}/data/",
            body={"format": "tsv-excel-noheader", "filterExpression": full_filter, "columns": columns_name},
        )

        schema = [schem for schem in dataset.get_schema()["columns"] if schem["name"] in columns_name]
        schema = list(sorted(schema, key=lambda x: columns_name.index(x["name"])))

        dataset_iterator = DataikuStreamedHttpUTF8CSVReader(schema=schema, csv_stream=csv_stream, chunk_size=limit)

        dataset_chunk = dataset_iterator.iter_rows(chunk_index=batch_offset)

        return self.return_dataset_chunk(dataset_chunk, format_, schema)

    def select_first_from_dataset_where_columns_are_values(
        self,
        dataset_name: str,
        filter_columns: List[str],
        filter_values: List[str],
        format_: Literal["dataframe", "iter"] = "dataframe",
    ) -> Union[pd.DataFrame, iter]:
        dataset = self.default_project.get_dataset(dataset_name)
        project_key = self.default_project.project_key

        all_filters = [
            f'toLowercase(toString(strval("{escape_string_double(c)}"))) == "{escape_string_double(str(filter_values[i])).lower()}"'
            for i, c in enumerate(filter_columns)
        ]

        full_filter = f"and({', '.join(all_filters)})"

        csv_stream = self.client._perform_raw(
            "POST",
            f"/projects/{project_key}/datasets/{dataset_name}/data/",
            body={
                "format": "tsv-excel-noheader",
                "filterExpression": full_filter,
                "sampling": {"samplingMethod": "HEAD_SEQUENTIAL"},
            },
        )

        schema = dataset.get_schema()["columns"]

        dataset_iterator = DataikuStreamedHttpUTF8CSVReader(schema=schema, csv_stream=csv_stream, chunk_size=1)

        dataset_chunk = dataset_iterator.iter_rows()

        return self.return_dataset_chunk(dataset_chunk, format_, schema)
    
    def return_dataset_chunk(self, dataset_chunk: iter, format_: str, schema: List[Dict]) -> Union[pd.DataFrame, iter]:
        if format_ == "dataframe":
            columns_name = [c["name"] for c in schema]
            return pd.DataFrame(columns=columns_name, data=next(dataset_chunk))
        else:
            try:
                chunk = next(dataset_chunk)
                return (r for r in chunk)
            except StopIteration:
                return iter([])

    def is_dataset_sql(self, dataset_name: str) -> bool:
        dataset = self.default_project.get_dataset(dataset_name)
        dataset_settings = dataset.get_settings().get_raw()

        dataset_type = dataset_settings["type"]
        is_dataset_sql = self.is_dataset_sql_supported(dataset_type)

        return is_dataset_sql

    def get_flow_zone_by_name(self, zone_name: str) -> Union[DSSFlowZone, None]:
        default_project_flow = self.default_project.get_flow()
        project_zones = default_project_flow.list_zones()
        legit_zones = [z for z in project_zones if z.name == zone_name]
        if len(legit_zones) == 0:
            logger.debug(f"No zones exist with the name {zone_name}.")
            return None

        return legit_zones[0]

    def create_flow_zone(self, zone_name: str) -> DSSFlowZone:
        default_project_flow = self.default_project.get_flow()
        new_zone = default_project_flow.create_zone(zone_name, color=ADVISOR_LIGHT_BLUE)
        return new_zone

    def get_or_create_flow_zone(self, zone_name: str) -> DSSFlowZone:
        existing_flow_zone = self.get_flow_zone_by_name(zone_name)
        if existing_flow_zone is None:
            existing_flow_zone = self.create_flow_zone(zone_name)
        return existing_flow_zone

    def list_managed_folders(self) -> List[Any]:
        return self.default_project.list_managed_folders()

    def get_managed_folder(self, id_: str) -> Union[dataiku.Folder, None]:
        managed_folder = dataiku.Folder(id_)
        try:
            _ = managed_folder.get_info()
            return managed_folder
        except Exception as err:
            logger.debug(f"Folder with id {id_} does probably not exist.")
            logger.exception(err)
            return None

    def get_managed_folder_by_name(self, name: str) -> Union[dataiku.Folder, None]:
        managed_folders = dataiku_api.list_managed_folders()
        matching_mfs = [mf for mf in managed_folders if mf["name"] == name]
        if len(matching_mfs) > 0:
            return self.get_managed_folder(matching_mfs[0]["id"])
        else:
            return None

    def create_empty_managed_folder(self, name: str, flow_zone_name=None) -> Union[dataiku.Folder, None]:
        try:
            new_managed_folder = self.default_project.create_managed_folder(name=name)
            if flow_zone_name:
                flow_zone = self.get_or_create_flow_zone(flow_zone_name)
                flow_zone.add_item(new_managed_folder)

            return dataiku.Folder(new_managed_folder.odb_id)
        except Exception as err:
            logger.debug(
                f"Something went wrong creating the managed folder. Are you sure you have a filesystem connection available?"
            )
            logger.exception(err)
            return None

    def get_or_create_managed_folder(
        self, managed_folder_name: str, flow_zone_name=None
    ) -> Union[dataiku.Folder, None]:
        existing_folder = self.get_managed_folder_by_name(managed_folder_name)
        if existing_folder is None:
            existing_folder = self.create_empty_managed_folder(managed_folder_name, flow_zone_name)
        return existing_folder


dataiku_api = DataikuApi()
