from dataiku.sql import Column, Constant, SelectQuery, toSQL
from dataiku.core.sql import SQLExecutor2, QueryReader
import dataiku
import functools
import pandas as pd
from typing import List, Union
from typing_extensions import Literal
from backend.utils.dataiku_api import dataiku_api
from backend.utils.misc import escape_string
from werkzeug.exceptions import BadRequest
import re
import logging

logger = logging.getLogger(__name__)


def execute(
    query_raw,
    dataset: str,
    format_: Literal["dataframe", "iter"] = "dataframe",
    offset: int = None,
    debug_query=False,
):
    try:
        query = toSQL(query_raw, dataset=dataset)
        if offset is not None:
            query = get_query_with_offset(query, offset)
        if debug_query:
            logger.info(f"Executing query: {query}")

    except Exception as err:
        raise BadRequest(f"Error when generating SQL query: {err}")

    sql_executor = SQLExecutor2(dataset=dataset)

    if format_ == "dataframe":
        try:
            query_result = sql_executor.query_to_df(query=query).fillna("")
        except Exception as err:
            raise BadRequest(f"Error when executing SQL query: {err}")
        return query_result
    elif format_ == "iter":
        try:
            query_result = sql_executor.query_to_iter(query=query).iter_tuples()
        except Exception as err:
            raise BadRequest(f"Error when executing SQL query: {err}")
        return query_result


def get_query_with_offset(query: str, offset: int):
    if not "LIMIT" in query:
        return query
    query = re.sub(r"(LIMIT [0-9]+)", rf"\1 OFFSET {offset}", query)
    return query


def _sql_true():
    return Constant(1).eq(Constant(1))


def _get_keep_none_cond(columns: List[Column], keep_none: Union[bool, List[bool]] = False):
    cond = _sql_true()
    if isinstance(keep_none, bool):
        if not keep_none:
            cond = functools.reduce(lambda c1, c2: c1.and_(c2.is_not_null()), columns, cond)
    elif isinstance(keep_none, list):
        if len(keep_none) != len(columns):
            raise Exception("keep_none should have the same length as columns_name.")
        for i, kn in enumerate(keep_none):
            if columns[i] and not kn:
                cond = cond.and_(columns[i].is_not_null())
    return cond


def _get_like_string_cond(columns: List[Column], like_string: Union[str, List[str]] = ""):
    cond = _sql_true()
    if isinstance(like_string, str):
        if len(like_string) == 0:
            return cond
        like_string = escape_string(like_string=like_string).lower()
        like_string = f"%{like_string}%"
        cond = functools.reduce(
            lambda c1, c2: c1.and_(c2.cast("STRING").lower().like(like_string)),
            columns,
            _sql_true(),
        )
    elif isinstance(like_string, list):
        if all([len(ls) == 0 for ls in like_string]):
            return cond
        if len(like_string) != len(columns):
            raise Exception("like_string should have the same length as columns_name.")
        for i, ls in enumerate(like_string):
            if len(ls):
                ls_ready = escape_string(like_string=ls).lower()
                like_string = f"%{ls_ready}%"
                cond = cond.and_(columns[i].cast("STRING").lower().like(ls_ready))
    return cond


def get_rows_count(dataset_name: str) -> int:
    select_query = SelectQuery()
    dataset = dataiku.Dataset(dataset_name, project_key=dataiku_api.default_project.project_key)
    select_query.select_from(dataset)
    select_query.select(Column("*").count())
    rows_count = execute(select_query, dataset, "iter")

    return next(rows_count)[0]


def select_columns_from_dataset(
    dataset_name: str,
    columns_name: List[str],
    format_: Literal["dataframe", "iter"] = "dataframe",
    distinct: bool = False,
    keep_none: Union[bool, List[bool]] = False,
    like_string: Union[str, List[str]] = "",
    limit: int = None,
    order_by: bool = False,
    batch_offset: int = 0,
) -> Union[pd.DataFrame, iter]:
    dataset = dataiku.Dataset(dataset_name, project_key=dataiku_api.default_project.project_key)

    columns_to_select = [Column(str(c)) for c in columns_name if c]

    select_query = SelectQuery()
    if distinct:
        select_query.distinct()
    select_query.select_from(dataset)

    select_query.select(columns_to_select)

    like_string_cond = _get_like_string_cond(columns=columns_to_select, like_string=like_string)
    keep_none_cond = _get_keep_none_cond(columns=columns_to_select, keep_none=keep_none)
    select_query.where(keep_none_cond.and_(like_string_cond))

    if order_by:
        select_query.order_by(columns_to_select[0])

    if limit:
        select_query.limit(limit)

    offset = None
    if batch_offset:
        offset = limit * batch_offset

    return execute(select_query, dataset, format_, offset=offset, debug_query=True)


def select_first_from_dataset_where_columns_are_values(
    dataset_name: str,
    filter_columns: List[str],
    filter_values: List[str],
    format_: Literal["dataframe", "iter"] = "dataframe",
) -> Union[pd.DataFrame, iter]:
    dataset = dataiku.Dataset(dataset_name, project_key=dataiku_api.default_project.project_key)
    select_query = SelectQuery()
    select_query.select_from(dataset)
    select_query.select(Column("*"))

    if len(filter_columns) != len(filter_values):
        raise Exception("filter_columns should have the same length as filter_values.")
    cond = _sql_true()
    for i, c in enumerate(filter_columns):
        eq_value = str(filter_values[i]).lower()
        if eq_value:
            cond = cond.and_(Column(c).cast("STRING").lower().eq(eq_value))

    select_query.where(cond)

    select_query.limit(1)
    return execute(select_query, dataset, format_)
